/*  alltel.c -- Linux kernel module to support the alltel parallel device
    Copyright (C) 2000  Shane Wegner

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    To contact the author, please write to Shane Wegner, 10425 140B st,
    Surrey BC, V3T 4N9 Canada or email shane@cm.nu.

   This module aimes to be a portable driver for the Alltel parallel
   telephone device.  This is my first attempt at a kernel module.  I
   imagine it'll be pretty ugly.

   TODO:
      * Add support for v2.0 kernels.

   */

/* $Id: alltel.c,v 1.24 2000/05/22 18:48:05 shane Exp $ */

#ifdef __sparc__
#error "This module cannot compile on the Sparc platform"
#endif

#include <linux/module.h>
#ifdef CONFIG_SMP
#define __SMP__
#endif
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/tqueue.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <asm/io.h>

#include "alltel.h"

#define ALLTEL_VERSION "0.11"
#define ALLTEL_LASTMOD "2000-05-21"
/* Number of structures to buffer for each device */
#define ALLTEL_BUFFER_EVENTS 10
/* nr of devices to support */
#define ALLTEL_MAX_DEVICES 4
/* Hardware specific mappings */
/* Mappings for the DTMF chip */
/* 1-9 are as expected */
#define ALLTEL_DTMF_D 0
#define ALLTEL_DTMF_0 10
#define ALLTEL_DTMF_STAR 11
#define ALLTEL_DTMF_POUND 12
#define ALLTEL_DTMF_A 13
#define ALLTEL_DTMF_B 14
#define ALLTEL_DTMF_C 15
#define ALLTEL_FLAG_LOOPBREAK 0x10 /* select pin 13 */
#define ALLTEL_FLAG_OFFHOOK 0x20 /* Paper end pin 12 */
#define ALLTEL_FLAG_DTMF 0x40 /* acknowledge pin 10 */
#define ALLTEL_FLAG_ISREADY 0x80 /* busy pin 11 */

int alltel_major = 0;
int alltel_use_count[ALLTEL_MAX_DEVICES] = {0, };
#ifdef __alpha__
int alltel_base[ALLTEL_MAX_DEVICES] = {0x3bc, };
#else
int alltel_base[ALLTEL_MAX_DEVICES] = {0x378, }; /* default to lp0 */
#endif
int alltel_irq[ALLTEL_MAX_DEVICES] = {-1, -1, -1, -1};
int alltel_use_irq[ALLTEL_MAX_DEVICES] = {0, };
unsigned int alltel_poll_time[ALLTEL_MAX_DEVICES] = {10, 10, 10, 10}; /* # of times per second */
unsigned int alltel_poll_time_notready[ALLTEL_MAX_DEVICES] = {1, 1, 1, 1}; /* Don't poll as often if the device is not ready */

static struct alltel_event *alltel_buffer[ALLTEL_MAX_DEVICES];
static volatile size_t alltel_head[ALLTEL_MAX_DEVICES] = {0, };
static struct wait_queue *alltel_queue[ALLTEL_MAX_DEVICES];
static struct timer_list alltel_timer[ALLTEL_MAX_DEVICES];
static unsigned int last_status[ALLTEL_MAX_DEVICES];
spinlock_t alltel_lock = SPIN_LOCK_UNLOCKED;

struct alltel_dev
{
volatile size_t alltel_tail;
};

MODULE_AUTHOR("Shane Wegner <shane@cm.nu>");
MODULE_SUPPORTED_DEVICE("alltel");
MODULE_PARM(alltel_major, "i");
MODULE_PARM(alltel_base, "1-" __MODULE_STRING(ALLTEL_MAX_DEVICES) "i");
MODULE_PARM(alltel_irq, "1-" __MODULE_STRING(ALLTEL_MAX_DEVICES) "i");
MODULE_PARM(alltel_use_irq, "1-" __MODULE_STRING(ALLTEL_MAX_DEVICES) "i");
MODULE_PARM(alltel_poll_time, "1-" __MODULE_STRING(ALLTEL_MAX_DEVICES) "i");
MODULE_PARM(alltel_poll_time_notready, "1-" __MODULE_STRING(ALLTEL_MAX_DEVICES) "i");

static void alltel_makeevent(unsigned int data, unsigned int status,
struct alltel_event *event)
{
#ifdef ALLTEL_DEBUG
printk(KERN_DEBUG "alltel: makeevent called: status=%x, data=%x\n",
status,data);
#endif
do_gettimeofday(&event->al_time);
if (status & ALLTEL_FLAG_ISREADY)
event->al_isready = 1;
else {
event->al_isready = 0;
status &= ~(ALLTEL_FLAG_DTMF | ALLTEL_FLAG_LOOPBREAK | ALLTEL_FLAG_OFFHOOK);
}
if (status & ALLTEL_FLAG_LOOPBREAK)
{
event->al_loopbreak = 1;
status &= ~ALLTEL_FLAG_DTMF;
}
else
event->al_loopbreak = 0;
if (status & ALLTEL_FLAG_OFFHOOK)
event->al_offhook = 1;
else
{
event->al_offhook = 0;
status &= ~ALLTEL_FLAG_DTMF;
}
if (status & ALLTEL_FLAG_DTMF)
{
event->al_dtmf = 1;
data &= 0x0f; /* High bits aren't used */
if (data >=1 && data <= 9)
event->al_digit = data + 48;
else
switch(data)
{
case ALLTEL_DTMF_0:
event->al_digit = '0';
break;
case ALLTEL_DTMF_STAR:
event->al_digit = '*';
break;
case ALLTEL_DTMF_POUND:
event->al_digit = '#';
break;
case ALLTEL_DTMF_A:
event->al_digit = 'A';
break;
case ALLTEL_DTMF_B:
event->al_digit = 'B';
break;
case ALLTEL_DTMF_C:
event->al_digit = 'C';
break;
case ALLTEL_DTMF_D:
event->al_digit = 'D';
break;
default:
event->al_dtmf = 0;
} /* switch */
} else /* al_dtmf */
event->al_dtmf = 0;
}

void alltel_do_timer(unsigned long arg)
{
unsigned int data, status;
status = inb(alltel_base[arg] + 1);
spin_lock(&alltel_lock);
if (status != last_status[arg])
{
last_status[arg] = status;
data = inb(alltel_base[arg]);
alltel_makeevent(data, status, &alltel_buffer[arg][alltel_head[arg]++]);
if (alltel_head[arg] >= ALLTEL_BUFFER_EVENTS)
alltel_head[arg] = 0;
wake_up_interruptible(&alltel_queue[arg]);
}
spin_unlock(&alltel_lock);
if (status & ALLTEL_FLAG_ISREADY)
alltel_timer[arg].expires = jiffies + (HZ / alltel_poll_time[arg]);
else
alltel_timer[arg].expires = jiffies + (HZ / alltel_poll_time_notready[arg]);
add_timer(&alltel_timer[arg]);
}

int alltel_open(struct inode *inode, struct file *filp)
{
struct alltel_dev *dev;
unsigned int i = MINOR(inode->i_rdev);
if (MAJOR(inode->i_rdev) != alltel_major)
return -EINVAL;
if (i >= ALLTEL_MAX_DEVICES || !alltel_base[i])
return -ENODEV;
MOD_INC_USE_COUNT;
alltel_use_count[i]++;
dev = kmalloc(sizeof(struct alltel_dev), GFP_KERNEL);
if (dev == NULL)
{
alltel_use_count[i]--;
MOD_DEC_USE_COUNT;
return -ENOMEM;
}
spin_lock(&alltel_lock);
dev->alltel_tail = alltel_head[i];
spin_unlock(&alltel_lock);
filp->private_data = dev;
if (alltel_use_count[i] == 1)
{
spin_lock(&alltel_lock);
last_status[i] = inb(alltel_base[i] +1 );
spin_unlock(&alltel_lock);
alltel_do_timer(i);
}
return 0;
}

int alltel_release(struct inode *inode, struct file *filp)
{
struct alltel_dev *dev = filp->private_data;
unsigned int i = MINOR(inode->i_rdev);
if (alltel_use_count[i] == 1)
del_timer(&alltel_timer[i]);
kfree(dev);
alltel_use_count[i]--;
MOD_DEC_USE_COUNT;
return 0;
}

ssize_t alltel_read(struct file *filp,
char *buf, size_t count, loff_t *ppos)
{
struct alltel_dev *dev = filp->private_data;
size_t retval = 0, i = 0;
unsigned int minor = MINOR(filp->f_dentry->d_inode->i_rdev);
if (!(unsigned long)(count / sizeof(struct alltel_event)))
return -EINVAL;
spin_lock(&alltel_lock);
while(alltel_head[minor] == dev->alltel_tail)
{
if (filp->f_flags & O_NONBLOCK)
retval = -EAGAIN;
if (signal_pending(current))
retval = -ERESTARTSYS;
spin_unlock(&alltel_lock);
if (retval)
return retval;
interruptible_sleep_on(&alltel_queue[minor]);
spin_lock(&alltel_lock);
} /* while */
while (alltel_head[minor] != dev->alltel_tail && i <= count - sizeof(struct alltel_event))
{
retval = copy_to_user(buf + i, &alltel_buffer[minor][dev->alltel_tail],
sizeof(struct alltel_event))
? -EFAULT : sizeof(struct alltel_event);
if (retval >= 0)
{
dev->alltel_tail++;
if (dev->alltel_tail >= ALLTEL_BUFFER_EVENTS)
dev->alltel_tail = 0;
i += retval;
} else {
i = retval;
break;
}
} /* while */
spin_unlock(&alltel_lock);
return i;
}

unsigned int alltel_poll(struct file *filp, poll_table *table)
{
struct alltel_dev *dev = filp->private_data;
int retval = 0;
unsigned int i = MINOR(filp->f_dentry->d_inode->i_rdev);
poll_wait(filp, &alltel_queue[i], table);
spin_lock(&alltel_lock);
if (alltel_head[i] != dev->alltel_tail)
retval = POLLIN | POLLRDNORM;
spin_unlock(&alltel_lock);
return retval;
}

int alltel_ioctl(struct inode *inode, struct file *filp,
unsigned int cmd, unsigned long arg)
{
int retval = 0;
struct alltel_event event;
unsigned int i = MINOR(inode->i_rdev);
switch(cmd)
{
case ALLTEL_IOCGSTATUS:
alltel_makeevent(inb(alltel_base[i]), inb(alltel_base[i]+1), &event);
retval = copy_to_user((struct alltel_event *)arg, &event,
sizeof(struct alltel_event)) ? -EFAULT : 0;
break;
default:
retval = -EINVAL;
}
return retval;
}

struct file_operations alltel_fops = {
open: alltel_open,
release: alltel_release,
read: alltel_read,
poll: alltel_poll,
ioctl: alltel_ioctl,
};

static struct tq_struct alltel_task[ALLTEL_MAX_DEVICES];
static struct {
unsigned int data;
unsigned int status;
unsigned int minor;
} alltel_bh_arg;

void alltel_interrupt_bh(void *unused)
{
unsigned int i = alltel_bh_arg.minor;
#ifdef ALLTEL_DEBUG
printk(KERN_DEBUG "alltel: alltel_interrupt_bh called\n");
#endif
spin_lock(&alltel_lock);
/* If we didn't detect a release, generate one anyway as obviously the poll was too slow */
if (last_status[i] & ALLTEL_FLAG_DTMF)
{
alltel_makeevent(0, last_status[i] & ~ALLTEL_FLAG_DTMF,
&alltel_buffer[i][alltel_head[i]++]);
if (alltel_head[i] >= ALLTEL_BUFFER_EVENTS)
alltel_head[i] = 0;
}
last_status[i] = alltel_bh_arg.status;
alltel_makeevent(alltel_bh_arg.data, alltel_bh_arg.status,
&alltel_buffer[i][alltel_head[i]++]);
if (alltel_head[i] >= ALLTEL_BUFFER_EVENTS)
alltel_head[i] = 0;
spin_unlock(&alltel_lock);
wake_up_interruptible(&alltel_queue[i]);
}

void alltel_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
unsigned int i;
for (i = 0; i < ALLTEL_MAX_DEVICES; i++)
if (irq == alltel_irq[i])
{
alltel_bh_arg.minor = i;
break;
}
if (i == ALLTEL_MAX_DEVICES)
printk(KERN_WARNING "alltel: Can't find irq %d in device list\n", irq);
else {
alltel_bh_arg.status = inb(alltel_base[i] + 1);
alltel_bh_arg.data = inb(alltel_base[i]);
queue_task(&alltel_task[i], &tq_immediate);
mark_bh(IMMEDIATE_BH);
}
}

static int alltel_init_device(unsigned int i)
{
int result;
if ((result = check_region(alltel_base[i], 4)))
{
printk(KERN_ERR "alltel%d: Can't get I/O address 0x%x-0x%x\n",
i, alltel_base[i], alltel_base[i] + 4);
return result;
}
if ((result = check_region(alltel_base[i] + 0x402, 1)))
{
printk(KERN_ERR "alltel%d: Can't get I/O address 0x%x\n",
i, alltel_base[i] + 0x402);
return result;
}
printk(KERN_INFO "alltel%d: base address set to 0x%x\n", i, alltel_base[i]);
request_region(alltel_base[i], 4, "alltel");
request_region(alltel_base[i] + 0x402, 1, "alltel");
alltel_buffer[i] = kmalloc(sizeof(struct alltel_event) * ALLTEL_BUFFER_EVENTS, GFP_KERNEL);
if (alltel_buffer[i] == NULL)
{
printk(KERN_ERR "alltel%d: Unable to allocate memory for read buffer", i);
release_region(alltel_base[i], 4);
release_region(alltel_base[i] + 0x402, 4);
return -ENOMEM;
}
/* check for an ecp parallel port */
if ((inb(alltel_base[i] + 0x402) & 0x03) == 1)
{
outb(0x34, alltel_base[i] + 0x402); /* set port to byte mode */
if (inb(alltel_base[i] + 0x402) == 0x35)
printk(KERN_INFO "alltel%d: ECP parallel port detected.\n", i);
}
outb(0x20, alltel_base[i]+2); /* enable bidirectional mode */
/* see if it's a bidirectional port */
outb(0xa0, alltel_base[i]);
if (inb(alltel_base[i]) == 0xa0)
{
/* check one more time */
outb(0xd0, alltel_base[i]);
if (inb(alltel_base[i]) == 0xd0)
printk(KERN_WARNING "alltel%d: this may not be a bidirectional parallel port.\n", i);
}
outb(0x00, alltel_base[i]);
if (alltel_use_irq[i] && alltel_irq[i] == -1)
switch(alltel_base[i])
{
case 0x378:
alltel_irq[i] = 7;
break;
case 0x278:
alltel_irq[i] = 2;
break;
case 0x3bc:
alltel_irq[i] = 5;
break;
default:
printk(KERN_WARNING "alltel%d: Don't know how to detect irq for 0x%x base\n",
i, alltel_base[i]);
alltel_use_irq[i] = 0;
} /* switch */
if (alltel_use_irq[i] && alltel_irq[i] >= 0)
{
alltel_task[i].routine = alltel_interrupt_bh;
alltel_task[i].data = NULL;
if (request_irq(alltel_irq[i], alltel_interrupt, SA_SHIRQ | SA_SAMPLE_RANDOM,
"alltel", NULL))
{
alltel_use_irq[i] = 0;
printk(KERN_WARNING "alltel%d: Unable to obtain requested irq.\n", i);
} else /* actually enable it */
outb(0x30, alltel_base[i]+2); /* enable irq and bidirectional mode */
} else
alltel_use_irq[i] = 0;
/* endif alltel_irq >= 0 */
if (alltel_use_irq[i])
{
printk(KERN_INFO "alltel%d: Interrupt driven operation enabled on irq %d\n",
i, alltel_irq[i]);
} else
printk(KERN_INFO "alltel%d: Interrupt driven operation disabled\n", i);
/* initialize our timer */
init_timer(&alltel_timer[i]);
alltel_timer[i].function = alltel_do_timer;
alltel_timer[i].data = i;
return 0;
}

int init_module(void)
{
unsigned int i;
int result;
printk(KERN_INFO "Alltel: v%s %s Shane WegneR <http://www.cm.nu/~shane/alltel/>\n",
ALLTEL_VERSION, ALLTEL_LASTMOD);
if ((result = register_chrdev(alltel_major, "alltel", &alltel_fops)) < 0)
{
printk(KERN_ERR "alltel: Can't get major number\n");
return result;
}
if (alltel_major == 0)
{
alltel_major = result; /* dynamic */
printk(KERN_INFO "alltel: Dynamic major number set to %d\n", alltel_major);
}
for (i = 0; alltel_base[i] != 0; i++)
{
result = alltel_init_device(i);
if (result && i)
{
alltel_base[i] = 0;
break;
} else if (result)
{
unregister_chrdev(alltel_major, "alltel");
return result;
}
}
printk(KERN_INFO "alltel: Driver loaded\n");
return 0;
}

void cleanup_module(void)
{
unsigned int i;
for (i = 0; alltel_base[i] != 0; i++)
{
del_timer(&alltel_timer[i]);
if (alltel_use_irq[i])
free_irq(alltel_irq[i], NULL);
kfree(alltel_buffer[i]);
release_region(alltel_base[i], 4);
release_region(alltel_base[i] + 0x402, 1);
}
unregister_chrdev(alltel_major, "alltel");
printk(KERN_INFO "alltel: Driver unloaded\n");
return;
}
